@@ -27,8 +27,12 @@ the corresponding CQL or Cassandra type strings.
 # for example), these classes would be a good place to tack on
 # .from_cql_literal() and .as_cql_literal() classmethods (or whatever).
 
+from __future__ import absolute_import  # to enable import io from stdlib
+from binascii import unhexlify
 import calendar
+from collections import namedtuple
 from decimal import Decimal
+import io
 import re
 import socket
 import time
@@ -133,6 +137,7 @@ def parse_casstype_args(typestring):
         elif tok == ')':
             types, names = args.pop()
             prev_types, prev_names = args[-1]
+
             prev_types[-1] = prev_types[-1].apply_parameters(types, names)
         else:
             types, names = args[-1]
@@ -214,7 +219,7 @@ class _CassandraType(object):
         return val
 
     @classmethod
-    def from_binary(cls, byts):
+    def from_binary(cls, byts, protocol_version):
         """
         Deserialize a bytestring into a value. See the deserialize() method
         for more information. This method differs in that if None or the empty
@@ -224,19 +229,19 @@ class _CassandraType(object):
             return None
         elif len(byts) == 0 and not cls.empty_binary_ok:
             return EMPTY if cls.support_empty_values else None
-        return cls.deserialize(byts)
+        return cls.deserialize(byts, protocol_version)
 
     @classmethod
-    def to_binary(cls, val):
+    def to_binary(cls, val, protocol_version):
         """
         Serialize a value into a bytestring. See the serialize() method for
         more information. This method differs in that if None is passed in,
         the result is the empty string.
         """
-        return b'' if val is None else cls.serialize(val)
+        return b'' if val is None else cls.serialize(val, protocol_version)
 
     @staticmethod
-    def deserialize(byts):
+    def deserialize(byts, protocol_version):
         """
         Given a bytestring, deserialize into a value according to the protocol
         for this type. Note that this does not create a new instance of this
@@ -246,7 +251,7 @@ class _CassandraType(object):
         return byts
 
     @staticmethod
-    def serialize(val):
+    def serialize(val, protocol_version):
         """
         Given a value appropriate for this class, serialize it according to the
         protocol for this type and return the corresponding bytestring.
@@ -294,8 +299,9 @@ class _CassandraType(object):
         if cls.num_subtypes != 'UNKNOWN' and len(subtypes) != cls.num_subtypes:
             raise ValueError("%s types require %d subtypes (%d given)"
                              % (cls.typename, cls.num_subtypes, len(subtypes)))
-        # newname = cls.cass_parameterized_type_with(subtypes).encode('utf8')
         newname = cls.cass_parameterized_type_with(subtypes)
+        if six.PY2 and isinstance(newname, unicode):
+            newname = newname.encode('utf-8')
         return type(newname, (cls,), {'subtypes': subtypes, 'cassname': cls.cassname})
 
     @classmethod
@@ -347,7 +353,7 @@ class BytesType(_CassandraType):
         return bytearray(val)
 
     @staticmethod
-    def serialize(val):
+    def serialize(val, protocol_version):
         return six.binary_type(val)
 
 
@@ -359,13 +365,13 @@ class DecimalType(_CassandraType):
         return Decimal(val)
 
     @staticmethod
-    def deserialize(byts):
+    def deserialize(byts, protocol_version):
         scale = int32_unpack(byts[:4])
         unscaled = varint_unpack(byts[4:])
         return Decimal('%de%d' % (unscaled, -scale))
 
     @staticmethod
-    def serialize(dec):
+    def serialize(dec, protocol_version):
         try:
             sign, digits, exponent = dec.as_tuple()
         except AttributeError:
@@ -382,11 +388,11 @@ class UUIDType(_CassandraType):
     typename = 'uuid'
 
     @staticmethod
-    def deserialize(byts):
+    def deserialize(byts, protocol_version):
         return UUID(bytes=byts)
 
     @staticmethod
-    def serialize(uuid):
+    def serialize(uuid, protocol_version):
         try:
             return uuid.bytes
         except AttributeError:
@@ -401,11 +407,11 @@ class BooleanType(_CassandraType):
         return bool(val)
 
     @staticmethod
-    def deserialize(byts):
+    def deserialize(byts, protocol_version):
         return bool(int8_unpack(byts))
 
     @staticmethod
-    def serialize(truth):
+    def serialize(truth, protocol_version):
         return int8_pack(truth)
 
 
@@ -419,11 +425,11 @@ else:
         empty_binary_ok = True
 
         @staticmethod
-        def deserialize(byts):
+        def deserialize(byts, protocol_version):
             return byts.decode('ascii')
 
         @staticmethod
-        def serialize(var):
+        def serialize(var, protocol_version):
             try:
                 return var.encode('ascii')
             except UnicodeDecodeError:
@@ -433,36 +439,61 @@ else:
 class FloatType(_CassandraType):
     typename = 'float'
 
-    deserialize = staticmethod(float_unpack)
-    serialize = staticmethod(float_pack)
+    @staticmethod
+    def deserialize(byts, protocol_version):
+        return float_unpack(byts)
+
+    @staticmethod
+    def serialize(byts, protocol_version):
+        return float_pack(byts)
 
 
 class DoubleType(_CassandraType):
     typename = 'double'
 
-    deserialize = staticmethod(double_unpack)
-    serialize = staticmethod(double_pack)
+    @staticmethod
+    def deserialize(byts, protocol_version):
+        return double_unpack(byts)
+
+    @staticmethod
+    def serialize(byts, protocol_version):
+        return double_pack(byts)
 
 
 class LongType(_CassandraType):
     typename = 'bigint'
 
-    deserialize = staticmethod(int64_unpack)
-    serialize = staticmethod(int64_pack)
+    @staticmethod
+    def deserialize(byts, protocol_version):
+        return int64_unpack(byts)
+
+    @staticmethod
+    def serialize(byts, protocol_version):
+        return int64_pack(byts)
 
 
 class Int32Type(_CassandraType):
     typename = 'int'
 
-    deserialize = staticmethod(int32_unpack)
-    serialize = staticmethod(int32_pack)
+    @staticmethod
+    def deserialize(byts, protocol_version):
+        return int32_unpack(byts)
+
+    @staticmethod
+    def serialize(byts, protocol_version):
+        return int32_pack(byts)
 
 
 class IntegerType(_CassandraType):
     typename = 'varint'
 
-    deserialize = staticmethod(varint_unpack)
-    serialize = staticmethod(varint_pack)
+    @staticmethod
+    def deserialize(byts, protocol_version):
+        return varint_unpack(byts)
+
+    @staticmethod
+    def serialize(byts, protocol_version):
+        return varint_pack(byts)
 
 
 have_ipv6_packing = hasattr(socket, 'inet_ntop')
@@ -475,7 +506,7 @@ class InetAddressType(_CassandraType):
     # inet_ntop and inet_pton aren't available on Windows
 
     @staticmethod
-    def deserialize(byts):
+    def deserialize(byts, protocol_version):
         if len(byts) == 16:
             if not have_ipv6_packing:
                 raise Exception(
@@ -485,7 +516,7 @@ class InetAddressType(_CassandraType):
             return socket.inet_ntoa(byts)
 
     @staticmethod
-    def serialize(addr):
+    def serialize(addr, protocol_version):
         if ':' in addr:
             fam = socket.AF_INET6
             if not have_ipv6_packing:
@@ -497,12 +528,9 @@ class InetAddressType(_CassandraType):
             return socket.inet_aton(addr)
 
 
-class CounterColumnType(_CassandraType):
+class CounterColumnType(LongType):
     typename = 'counter'
 
-    deserialize = staticmethod(int64_unpack)
-    serialize = staticmethod(int64_pack)
-
 
 cql_time_formats = (
     '%Y-%m-%d %H:%M',
@@ -544,11 +572,11 @@ class DateType(_CassandraType):
         return self.val
 
     @staticmethod
-    def deserialize(byts):
+    def deserialize(byts, protocol_version):
         return datetime.utcfromtimestamp(int64_unpack(byts) / 1000.0)
 
     @staticmethod
-    def serialize(v):
+    def serialize(v, protocol_version):
         global _have_warned_about_timestamps
         try:
             converted = calendar.timegm(v.utctimetuple())
@@ -560,7 +588,8 @@ class DateType(_CassandraType):
 
             if not _have_warned_about_timestamps:
                 _have_warned_about_timestamps = True
-                warnings.warn("timestamp columns in Cassandra hold a number of "
+                warnings.warn(
+                    "timestamp columns in Cassandra hold a number of "
                     "milliseconds since the unix epoch.  Currently, when executing "
                     "prepared statements, this driver multiplies timestamp "
                     "values by 1000 so that the result of time.time() "
@@ -587,11 +616,11 @@ class TimeUUIDType(DateType):
         return unix_time_from_uuid1(self.val)
 
     @staticmethod
-    def deserialize(byts):
+    def deserialize(byts, protocol_version):
         return UUID(bytes=byts)
 
     @staticmethod
-    def serialize(timeuuid):
+    def serialize(timeuuid, protocol_version):
         try:
             return timeuuid.bytes
         except AttributeError:
@@ -603,11 +632,11 @@ class UTF8Type(_CassandraType):
     empty_binary_ok = True
 
     @staticmethod
-    def deserialize(byts):
+    def deserialize(byts, protocol_version):
         return byts.decode('utf8')
 
     @staticmethod
-    def serialize(ustr):
+    def serialize(ustr, protocol_version):
         try:
             return ustr.encode('utf-8')
         except UnicodeDecodeError:
@@ -626,18 +655,18 @@ class _ParameterizedType(_CassandraType):
         _CassandraType.__init__(self, val)
 
     @classmethod
-    def deserialize(cls, byts):
+    def deserialize(cls, byts, protocol_version):
         if not cls.subtypes:
             raise NotImplementedError("can't deserialize unparameterized %s"
                                       % cls.typename)
-        return cls.deserialize_safe(byts)
+        return cls.deserialize_safe(byts, protocol_version)
 
     @classmethod
-    def serialize(cls, val):
+    def serialize(cls, val, protocol_version):
         if not cls.subtypes:
             raise NotImplementedError("can't serialize unparameterized %s"
                                       % cls.typename)
-        return cls.serialize_safe(val)
+        return cls.serialize_safe(val, protocol_version)
 
 
 class _SimpleParameterizedType(_ParameterizedType):
@@ -647,30 +676,37 @@ class _SimpleParameterizedType(_ParameterizedType):
         return cls.adapter([subtype.validate(subval) for subval in val])
 
     @classmethod
-    def deserialize_safe(cls, byts):
+    def deserialize_safe(cls, byts, protocol_version):
         subtype, = cls.subtypes
-        numelements = uint16_unpack(byts[:2])
-        p = 2
+        if protocol_version >= 3:
+            unpack = int32_unpack
+            length = 4
+        else:
+            unpack = uint16_unpack
+            length = 2
+        numelements = unpack(byts[:length])
+        p = length
         result = []
         for _ in range(numelements):
-            itemlen = uint16_unpack(byts[p:p + 2])
-            p += 2
+            itemlen = unpack(byts[p:p + length])
+            p += length
             item = byts[p:p + itemlen]
             p += itemlen
-            result.append(subtype.from_binary(item))
+            result.append(subtype.from_binary(item, protocol_version))
         return cls.adapter(result)
 
     @classmethod
-    def serialize_safe(cls, items):
+    def serialize_safe(cls, items, protocol_version):
         if isinstance(items, six.string_types):
             raise TypeError("Received a string for a type that expects a sequence")
 
         subtype, = cls.subtypes
-        buf = six.BytesIO()
-        buf.write(uint16_pack(len(items)))
+        pack = int32_pack if protocol_version >= 3 else uint16_pack
+        buf = io.BytesIO()
+        buf.write(pack(len(items)))
         for item in items:
-            itembytes = subtype.to_binary(item)
-            buf.write(uint16_pack(len(itembytes)))
+            itembytes = subtype.to_binary(item, protocol_version)
+            buf.write(pack(len(itembytes)))
             buf.write(itembytes)
         return buf.getvalue()
 
@@ -678,7 +714,7 @@ class _SimpleParameterizedType(_ParameterizedType):
 class ListType(_SimpleParameterizedType):
     typename = 'list'
     num_subtypes = 1
-    adapter = tuple
+    adapter = list
 
 
 class SetType(_SimpleParameterizedType):
@@ -697,52 +733,197 @@ class MapType(_ParameterizedType):
         return dict((subkeytype.validate(k), subvaltype.validate(v)) for (k, v) in six.iteritems(val))
 
     @classmethod
-    def deserialize_safe(cls, byts):
+    def deserialize_safe(cls, byts, protocol_version):
         subkeytype, subvaltype = cls.subtypes
-        numelements = uint16_unpack(byts[:2])
-        p = 2
+        if protocol_version >= 3:
+            unpack = int32_unpack
+            length = 4
+        else:
+            unpack = uint16_unpack
+            length = 2
+        numelements = unpack(byts[:length])
+        p = length
         themap = OrderedDict()
         for _ in range(numelements):
-            key_len = uint16_unpack(byts[p:p + 2])
-            p += 2
+            key_len = unpack(byts[p:p + length])
+            p += length
             keybytes = byts[p:p + key_len]
             p += key_len
-            val_len = uint16_unpack(byts[p:p + 2])
-            p += 2
+            val_len = unpack(byts[p:p + length])
+            p += length
             valbytes = byts[p:p + val_len]
             p += val_len
-            key = subkeytype.from_binary(keybytes)
-            val = subvaltype.from_binary(valbytes)
+            key = subkeytype.from_binary(keybytes, protocol_version)
+            val = subvaltype.from_binary(valbytes, protocol_version)
             themap[key] = val
         return themap
 
     @classmethod
-    def serialize_safe(cls, themap):
+    def serialize_safe(cls, themap, protocol_version):
         subkeytype, subvaltype = cls.subtypes
-        buf = six.BytesIO()
-        buf.write(uint16_pack(len(themap)))
+        pack = int32_pack if protocol_version >= 3 else uint16_pack
+        buf = io.BytesIO()
+        buf.write(pack(len(themap)))
         try:
             items = six.iteritems(themap)
         except AttributeError:
             raise TypeError("Got a non-map object for a map value")
         for key, val in items:
-            keybytes = subkeytype.to_binary(key)
-            valbytes = subvaltype.to_binary(val)
-            buf.write(uint16_pack(len(keybytes)))
+            keybytes = subkeytype.to_binary(key, protocol_version)
+            valbytes = subvaltype.to_binary(val, protocol_version)
+            buf.write(pack(len(keybytes)))
             buf.write(keybytes)
-            buf.write(uint16_pack(len(valbytes)))
+            buf.write(pack(len(valbytes)))
             buf.write(valbytes)
         return buf.getvalue()
 
 
+class TupleType(_ParameterizedType):
+    typename = 'tuple'
+    num_subtypes = 'UNKNOWN'
+
+    @classmethod
+    def deserialize_safe(cls, byts, protocol_version):
+        proto_version = max(3, protocol_version)
+        p = 0
+        values = []
+        for col_type in cls.subtypes:
+            if p == len(byts):
+                break
+            itemlen = int32_unpack(byts[p:p + 4])
+            p += 4
+            if itemlen >= 0:
+                item = byts[p:p + itemlen]
+                p += itemlen
+            else:
+                item = None
+            # collections inside UDTs are always encoded with at least the
+            # version 3 format
+            values.append(col_type.from_binary(item, proto_version))
+
+        if len(values) < len(cls.subtypes):
+            nones = [None] * (len(cls.subtypes) - len(values))
+            values = values + nones
+
+        return tuple(values)
+
+    @classmethod
+    def serialize_safe(cls, val, protocol_version):
+        if len(val) > len(cls.subtypes):
+            raise ValueError("Expected %d items in a tuple, but got %d: %s" %
+                             (len(cls.subtypes), len(val), val))
+
+        proto_version = max(3, protocol_version)
+        buf = io.BytesIO()
+        for item, subtype in zip(val, cls.subtypes):
+            if item is not None:
+                packed_item = subtype.to_binary(item, proto_version)
+                buf.write(int32_pack(len(packed_item)))
+                buf.write(packed_item)
+            else:
+                buf.write(int32_pack(-1))
+        return buf.getvalue()
+
+
+class UserType(TupleType):
+    typename = "'org.apache.cassandra.db.marshal.UserType'"
+
+    _cache = {}
+
+    @classmethod
+    def make_udt_class(cls, keyspace, udt_name, names_and_types, mapped_class):
+        if six.PY2 and isinstance(udt_name, unicode):
+            udt_name = udt_name.encode('utf-8')
+
+        try:
+            return cls._cache[(keyspace, udt_name)]
+        except KeyError:
+            fieldnames, types = zip(*names_and_types)
+            instance = type(udt_name, (cls,), {'subtypes': types,
+                                               'cassname': cls.cassname,
+                                               'typename': udt_name,
+                                               'fieldnames': fieldnames,
+                                               'keyspace': keyspace,
+                                               'mapped_class': mapped_class})
+            cls._cache[(keyspace, udt_name)] = instance
+            return instance
+
+    @classmethod
+    def apply_parameters(cls, subtypes, names):
+        keyspace = subtypes[0]
+        udt_name = unhexlify(subtypes[1].cassname)
+        field_names = [unhexlify(encoded_name) for encoded_name in names[2:]]
+        assert len(field_names) == len(subtypes[2:])
+        return type(udt_name, (cls,), {'subtypes': subtypes[2:],
+                                       'cassname': cls.cassname,
+                                       'typename': udt_name,
+                                       'fieldnames': field_names,
+                                       'keyspace': keyspace,
+                                       'mapped_class': None})
+
+    @classmethod
+    def cql_parameterized_type(cls):
+        return cls.typename
+
+    @classmethod
+    def deserialize_safe(cls, byts, protocol_version):
+        proto_version = max(3, protocol_version)
+        p = 0
+        values = []
+        for col_type in cls.subtypes:
+            if p == len(byts):
+                break
+            itemlen = int32_unpack(byts[p:p + 4])
+            p += 4
+            if itemlen >= 0:
+                item = byts[p:p + itemlen]
+                p += itemlen
+            else:
+                item = None
+            # collections inside UDTs are always encoded with at least the
+            # version 3 format
+            values.append(col_type.from_binary(item, proto_version))
+
+        if len(values) < len(cls.subtypes):
+            nones = [None] * (len(cls.subtypes) - len(values))
+            values = values + nones
+
+        if cls.mapped_class:
+            return cls.mapped_class(**dict(zip(cls.fieldnames, values)))
+        else:
+            Result = namedtuple(cls.typename, cls.fieldnames)
+            return Result(*values)
+
+    @classmethod
+    def serialize_safe(cls, val, protocol_version):
+        proto_version = max(3, protocol_version)
+        buf = io.BytesIO()
+        for fieldname, subtype in zip(cls.fieldnames, cls.subtypes):
+            item = getattr(val, fieldname)
+            if item is not None:
+                packed_item = subtype.to_binary(getattr(val, fieldname), proto_version)
+                buf.write(int32_pack(len(packed_item)))
+                buf.write(packed_item)
+            else:
+                buf.write(int32_pack(-1))
+        return buf.getvalue()
+
+
 class CompositeType(_ParameterizedType):
     typename = "'org.apache.cassandra.db.marshal.CompositeType'"
     num_subtypes = 'UNKNOWN'
 
+    @classmethod
+    def cql_parameterized_type(cls):
+        """
+        There is no CQL notation for Composites, so we override this.
+        """
+        typestring = cls.cass_parameterized_type(full=True)
+        return "'%s'" % (typestring,)
 
-class DynamicCompositeType(_ParameterizedType):
+
+class DynamicCompositeType(CompositeType):
     typename = "'org.apache.cassandra.db.marshal.DynamicCompositeType'"
-    num_subtypes = 'UNKNOWN'
 
 
 class ColumnToCollectionType(_ParameterizedType):
@@ -760,14 +941,14 @@ class ReversedType(_ParameterizedType):
     num_subtypes = 1
 
     @classmethod
-    def deserialize_safe(cls, byts):
+    def deserialize_safe(cls, byts, protocol_version):
         subtype, = cls.subtypes
         return subtype.from_binary(byts)
 
     @classmethod
-    def serialize_safe(cls, val):
+    def serialize_safe(cls, val, protocol_version):
         subtype, = cls.subtypes
-        return subtype.to_binary(val)
+        return subtype.to_binary(val, protocol_version)
 
 
 def is_counter_type(t):
